@@ -27,34 +27,41 @@
 from unicodedata import normalize
 import hashlib
 import re
-from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple
+import copy
+from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple, Any, Type
 from functools import lru_cache, wraps
 from abc import ABC, abstractmethod
 
-from . import bitcoin, ecc, constants, bip32
+import electrum_ecc as ecc
+from electrum_ecc import string_to_number
+
+from . import bitcoin, constants, bip32
 from .bitcoin import deserialize_privkey, serialize_privkey, BaseDecodeError
 from .transaction import Transaction, PartialTransaction, PartialTxInput, PartialTxOutput, TxInput
-from .bip32 import (convert_bip32_path_to_list_of_uint32, BIP32_PRIME,
+from .bip32 import (convert_bip32_strpath_to_intpath, BIP32_PRIME,
                     is_xpub, is_xprv, BIP32Node, normalize_bip32_derivation,
-                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info)
-from .ecc import string_to_number
+                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info,
+                    KeyOriginInfo)
+from .descriptor import PubkeyProvider
+from . import crypto
 from .crypto import (pw_decode, pw_encode, sha256, sha256d, PW_HASH_VERSION_LATEST,
                      SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion, hash_160,
                      CiphertextFormatError)
 from .util import (InvalidPassword, WalletFileException,
-                   BitcoinException, bh2u, bfh, inv_dict, is_hex_str)
-from .mnemonic import Mnemonic, Wordlist, seed_type, is_seed
+                   BitcoinException, bfh, inv_dict, is_hex_str)
+from .mnemonic import Mnemonic, Wordlist, calc_seed_type, is_seed
 from .plugin import run_hook
 from .logging import Logger
 
 if TYPE_CHECKING:
-    from .gui.qt.util import TaskThread
-    from .plugins.hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
+    from .gui.common_qt.util import TaskThread
+    from .hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
     from .wallet_db import WalletDB
     from .plugin import Device
 
 
 class CannotDerivePubkey(Exception): pass
+class ScriptTypeNotSupported(Exception): pass
 
 
 def also_test_none_password(check_password_fn):
@@ -98,31 +105,31 @@ class KeyStore(Logger, ABC):
         return f'{self.type}'
 
     @abstractmethod
-    def may_have_password(self):
+    def may_have_password(self) -> bool:
         """Returns whether the keystore can be encrypted with a password."""
         pass
 
-    def _get_tx_derivations(self, tx: 'PartialTransaction') -> Dict[str, Union[Sequence[int], str]]:
+    def _get_tx_derivations(self, tx: 'PartialTransaction') -> Dict[bytes, Union[Sequence[int], str]]:
         keypairs = {}
         for txin in tx.inputs():
             keypairs.update(self._get_txin_derivations(txin))
         return keypairs
 
-    def _get_txin_derivations(self, txin: 'PartialTxInput') -> Dict[str, Union[Sequence[int], str]]:
+    def _get_txin_derivations(self, txin: 'PartialTxInput') -> Dict[bytes, Union[Sequence[int], str]]:
         if txin.is_complete():
             return {}
         keypairs = {}
         for pubkey in txin.pubkeys:
-            if pubkey in txin.part_sigs:
+            if pubkey in txin.sigs_ecdsa:
                 # this pubkey already signed
                 continue
             derivation = self.get_pubkey_derivation(pubkey, txin)
             if not derivation:
                 continue
-            keypairs[pubkey.hex()] = derivation
+            keypairs[pubkey] = derivation
         return keypairs
 
-    def can_sign(self, tx: 'Transaction', *, ignore_watching_only=False) -> bool:
+    def can_sign(self, tx: 'Transaction', *, ignore_watching_only: bool = False) -> bool:
         """Returns whether this keystore could sign *something* in this tx."""
         if not ignore_watching_only and self.is_watching_only():
             return False
@@ -130,7 +137,7 @@ class KeyStore(Logger, ABC):
             return False
         return bool(self._get_tx_derivations(tx))
 
-    def can_sign_txin(self, txin: 'TxInput', *, ignore_watching_only=False) -> bool:
+    def can_sign_txin(self, txin: 'TxInput', *, ignore_watching_only: bool = False) -> bool:
         """Returns whether this keystore could sign this txin."""
         if not ignore_watching_only and self.is_watching_only():
             return False
@@ -142,7 +149,7 @@ class KeyStore(Logger, ABC):
         return not self.is_watching_only()
 
     @abstractmethod
-    def dump(self) -> dict:
+    def dump(self) -> dict[str, Any]:
         pass
 
     @abstractmethod
@@ -179,6 +186,10 @@ class KeyStore(Logger, ABC):
         """
         pass
 
+    @abstractmethod
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        pass
+
     def find_my_pubkey_in_txinout(
             self, txinout: Union['PartialTxInput', 'PartialTxOutput'],
             *, only_der_suffix: bool = False
@@ -193,10 +204,16 @@ class KeyStore(Logger, ABC):
     def can_have_deterministic_lightning_xprv(self) -> bool:
         return False
 
+    def has_support_for_slip_19_ownership_proofs(self) -> bool:
+        return False
+
+    def add_slip_19_ownership_proofs_to_tx(self, tx: 'PartialTransaction', *, password) -> None:
+        raise NotImplementedError()
+
 
 class Software_KeyStore(KeyStore):
 
-    def __init__(self, d):
+    def __init__(self, d: dict):
         KeyStore.__init__(self)
         self.pw_hash_version = d.get('pw_hash_version', 1)
         if self.pw_hash_version not in SUPPORTED_PW_HASH_VERSIONS:
@@ -208,12 +225,12 @@ class Software_KeyStore(KeyStore):
     def sign_message(self, sequence, message, password, *, script_type=None) -> bytes:
         privkey, compressed = self.get_private_key(sequence, password)
         key = ecc.ECPrivkey(privkey)
-        return key.sign_message(message, compressed)
+        return bitcoin.ecdsa_sign_usermessage(key, message, is_compressed=compressed)
 
     def decrypt_message(self, sequence, message, password) -> bytes:
         privkey, compressed = self.get_private_key(sequence, password)
         ec = ecc.ECPrivkey(privkey)
-        decrypted = ec.decrypt_message(message)
+        decrypted = crypto.ecies_decrypt_message(ec, message)
         return decrypted
 
     def sign_transaction(self, tx, password):
@@ -222,15 +239,17 @@ class Software_KeyStore(KeyStore):
         # Raise if password is not correct.
         self.check_password(password)
         # Add private keys
-        keypairs = self._get_tx_derivations(tx)
-        for k, v in keypairs.items():
-            keypairs[k] = self.get_private_key(v, password)
+        keypairs = {}
+        pubkey_to_deriv_map = self._get_tx_derivations(tx)
+        for pubkey, deriv in pubkey_to_deriv_map.items():
+            privkey, is_compressed = self.get_private_key(deriv, password)
+            keypairs[pubkey] = privkey
         # Sign
         if keypairs:
             tx.sign(keypairs)
 
     @abstractmethod
-    def update_password(self, old_password, new_password):
+    def update_password(self, old_password, new_password) -> None:
         pass
 
     @abstractmethod
@@ -249,7 +268,7 @@ class Imported_KeyStore(Software_KeyStore):
 
     type = 'imported'
 
-    def __init__(self, d):
+    def __init__(self, d: dict):
         Software_KeyStore.__init__(self, d)
         self.keypairs = d.get('keypairs', {})  # type: Dict[str, str]
 
@@ -271,7 +290,7 @@ class Imported_KeyStore(Software_KeyStore):
         pubkey = list(self.keypairs.keys())[0]
         self.get_private_key(pubkey, password)
 
-    def import_privkey(self, sec, password):
+    def import_privkey(self, sec: str, password) -> Tuple[str, str]:
         txin_type, privkey, compressed = deserialize_privkey(sec)
         pubkey = ecc.ECPrivkey(privkey).get_public_key_hex(compressed=compressed)
         # re-serialize the key so the internal storage format is consistent
@@ -284,7 +303,7 @@ class Imported_KeyStore(Software_KeyStore):
         self.keypairs[pubkey] = pw_encode(serialized_privkey, password, version=self.pw_hash_version)
         return txin_type, pubkey
 
-    def delete_imported_key(self, key):
+    def delete_imported_key(self, key: str) -> None:
         self.keypairs.pop(key)
 
     def get_private_key(self, pubkey: str, password):
@@ -302,6 +321,15 @@ class Imported_KeyStore(Software_KeyStore):
             return pubkey.hex()
         return None
 
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        if sequence in self.keypairs:
+            return PubkeyProvider(
+                origin=None,
+                pubkey=sequence,
+                deriv_path=None,
+            )
+        return None
+
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
         if new_password == '':
@@ -315,7 +343,7 @@ class Imported_KeyStore(Software_KeyStore):
 
 class Deterministic_KeyStore(Software_KeyStore):
 
-    def __init__(self, d):
+    def __init__(self, d: dict):
         Software_KeyStore.__init__(self, d)
         self.seed = d.get('seed', '')  # only electrum seeds
         self.passphrase = d.get('passphrase', '')
@@ -350,18 +378,18 @@ class Deterministic_KeyStore(Software_KeyStore):
     def format_seed(self, seed: str) -> str:
         pass
 
-    def add_seed(self, seed):
+    def add_seed(self, seed: str) -> None:
         if self.seed:
             raise Exception("a seed exists")
         self.seed = self.format_seed(seed)
-        self._seed_type = seed_type(seed) or None
+        self._seed_type = calc_seed_type(seed) or None
 
-    def get_seed(self, password):
+    def get_seed(self, password) -> str:
         if not self.has_seed():
             raise Exception("This wallet has no seed words")
         return pw_decode(self.seed, password, version=self.pw_hash_version)
 
-    def get_passphrase(self, password):
+    def get_passphrase(self, password) -> str:
         if self.passphrase:
             return pw_decode(self.passphrase, password, version=self.pw_hash_version)
         else:
@@ -403,6 +431,9 @@ class MasterPublicKeyMixin(ABC):
         """
         pass
 
+    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
+        return None
+
     @abstractmethod
     def derive_pubkey(self, for_change: int, n: int) -> bytes:
         """Returns pubkey at given path.
@@ -436,7 +467,7 @@ class MasterPublicKeyMixin(ABC):
         # 1. try fp against our root
         ks_root_fingerprint_hex = self.get_root_fingerprint()
         ks_der_prefix_str = self.get_derivation_prefix()
-        ks_der_prefix = convert_bip32_path_to_list_of_uint32(ks_der_prefix_str) if ks_der_prefix_str else None
+        ks_der_prefix = convert_bip32_strpath_to_intpath(ks_der_prefix_str) if ks_der_prefix_str else None
         if (ks_root_fingerprint_hex is not None and ks_der_prefix is not None and
                 fp_found.hex() == ks_root_fingerprint_hex):
             if path_found[:len(ks_der_prefix)] == ks_der_prefix:
@@ -490,7 +521,9 @@ class Xpub(MasterPublicKeyMixin):
         return self._xpub_bip32_node
 
     def get_derivation_prefix(self) -> Optional[str]:
-        return self._derivation_prefix
+        if self._derivation_prefix is None:
+            return None
+        return normalize_bip32_derivation(self._derivation_prefix)
 
     def get_root_fingerprint(self) -> Optional[str]:
         return self._root_fingerprint
@@ -506,11 +539,11 @@ class Xpub(MasterPublicKeyMixin):
         if not only_der_suffix and fingerprint_hex is not None and der_prefix_str is not None:
             # use root fp, and true full path
             fingerprint_bytes = bfh(fingerprint_hex)
-            der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+            der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
         else:
             # use intermediate fp, and claim der suffix is the full path
             fingerprint_bytes = self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()
-            der_prefix_ints = convert_bip32_path_to_list_of_uint32('m')
+            der_prefix_ints = convert_bip32_strpath_to_intpath('m')
         der_full = der_prefix_ints + list(der_suffix)
         return fingerprint_bytes, der_full
 
@@ -532,7 +565,23 @@ class Xpub(MasterPublicKeyMixin):
         )
         return bip32node.to_xpub()
 
-    def add_key_origin_from_root_node(self, *, derivation_prefix: str, root_node: BIP32Node):
+    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
+        fp_bytes, der_full = self.get_fp_and_derivation_to_be_used_in_partial_tx(
+            der_suffix=[], only_der_suffix=False)
+        origin = KeyOriginInfo(fingerprint=fp_bytes, path=der_full)
+        return origin
+
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        strpath = convert_bip32_intpath_to_strpath(sequence)
+        strpath = strpath[1:]  # cut leading "m"
+        bip32node = self.get_bip32_node_for_xpub()
+        return PubkeyProvider(
+            origin=self.get_key_origin_info(),
+            pubkey=bip32node._replace(xtype="standard").to_xkey(),
+            deriv_path=strpath,
+        )
+
+    def add_key_origin_from_root_node(self, *, derivation_prefix: str, root_node: BIP32Node) -> None:
         assert self.xpub
         # try to derive ourselves from what we were given
         child_node1 = root_node.subkey_at_private_derivation(derivation_prefix)
@@ -584,12 +633,19 @@ class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
 
     type = 'bip32'
 
-    def __init__(self, d):
+    def __init__(self, d: dict):
         Xpub.__init__(self, derivation_prefix=d.get('derivation'), root_fingerprint=d.get('root_fingerprint'))
         Deterministic_KeyStore.__init__(self, d)
         self.xpub = d.get('xpub')
         self.xprv = d.get('xprv')
 
+    def watching_only_keystore(self):
+        return BIP32_KeyStore({
+            'xpub': self.xpub,
+            'root_fingerprint': self.get_root_fingerprint(),
+            'derivation': self.get_derivation_prefix(),
+        })
+
     def format_seed(self, seed):
         return ' '.join(seed.split())
 
@@ -601,7 +657,7 @@ class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
         d['root_fingerprint'] = self.get_root_fingerprint()
         return d
 
-    def get_master_private_key(self, password):
+    def get_master_private_key(self, password) -> str:
         return pw_decode(self.xprv, password, version=self.pw_hash_version)
 
     @also_test_none_password
@@ -632,18 +688,18 @@ class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
     def is_watching_only(self):
         return self.xprv is None
 
-    def add_xpub(self, xpub):
+    def add_xpub(self, xpub: str) -> None:
         assert is_xpub(xpub)
         self.xpub = xpub
         root_fingerprint, derivation_prefix = bip32.root_fp_and_der_prefix_from_xkey(xpub)
         self.add_key_origin(derivation_prefix=derivation_prefix, root_fingerprint=root_fingerprint)
 
-    def add_xprv(self, xprv):
+    def add_xprv(self, xprv: str) -> None:
         assert is_xprv(xprv)
         self.xprv = xprv
         self.add_xpub(bip32.xpub_from_xprv(xprv))
 
-    def add_xprv_from_seed(self, bip32_seed, xtype, derivation):
+    def add_xprv_from_seed(self, bip32_seed: bytes, *, xtype: str, derivation: str) -> None:
         rootnode = BIP32Node.from_rootseed(bip32_seed, xtype=xtype)
         node = rootnode.subkey_at_private_derivation(derivation)
         self.add_xprv(node.to_xprv())
@@ -655,11 +711,6 @@ class BIP32_KeyStore(Xpub, Deterministic_KeyStore):
         pk = node.eckey.get_secret_bytes()
         return pk, True
 
-    def get_keypair(self, sequence, password):
-        k, _ = self.get_private_key(sequence, password)
-        cK = ecc.ECPrivkey(k).get_public_key_bytes()
-        return cK, k
-
     def can_have_deterministic_lightning_xprv(self):
         if (self.get_seed_type() == 'segwit'
                 and self.get_bip32_node_for_xpub().xtype == 'p2wpkh'):
@@ -677,25 +728,32 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
 
     type = 'old'
 
-    def __init__(self, d):
+    def __init__(self, d: dict):
         Deterministic_KeyStore.__init__(self, d)
-        self.mpk = d.get('mpk')
+        self.mpk = d.get('mpk')  # type: Optional[str]
         self._root_fingerprint = None
 
-    def get_hex_seed(self, password):
-        return pw_decode(self.seed, password, version=self.pw_hash_version).encode('utf8')
+    def watching_only_keystore(self):
+        return Old_KeyStore({'mpk': self.mpk})
+
+    def _get_hex_seed(self, password) -> str:
+        if not is_hex_str(self.seed) and password is None:
+            raise InvalidPassword()
+        hex_str = pw_decode(self.seed, password, version=self.pw_hash_version)
+        assert is_hex_str(hex_str), f"expected hex str, got {type(hex_str)} with {len(hex_str)=}"
+        return hex_str
 
     def dump(self):
         d = Deterministic_KeyStore.dump(self)
         d['mpk'] = self.mpk
         return d
 
-    def add_seed(self, seedphrase):
-        Deterministic_KeyStore.add_seed(self, seedphrase)
-        s = self.get_hex_seed(None)
-        self.mpk = self.mpk_from_seed(s)
+    def add_seed(self, seed):
+        Deterministic_KeyStore.add_seed(self, seed)
+        hex_seed = self._get_hex_seed(None)
+        self.mpk = self.mpk_from_seed(hex_seed)
 
-    def add_master_public_key(self, mpk):
+    def add_master_public_key(self, mpk: str) -> None:
         self.mpk = mpk
 
     def format_seed(self, seed):
@@ -716,28 +774,30 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
 
     def get_seed(self, password):
         from . import old_mnemonic
-        s = self.get_hex_seed(password)
-        return ' '.join(old_mnemonic.mn_encode(s))
+        hex_seed = self._get_hex_seed(password)
+        return ' '.join(old_mnemonic.mn_encode(hex_seed))
 
     @classmethod
-    def mpk_from_seed(klass, seed):
-        secexp = klass.stretch_key(seed)
+    def mpk_from_seed(cls, hex_seed: str) -> str:
+        secexp = cls.stretch_key(hex_seed)
         privkey = ecc.ECPrivkey.from_secret_scalar(secexp)
         return privkey.get_public_key_hex(compressed=False)[2:]
 
     @classmethod
-    def stretch_key(self, seed):
-        x = seed
+    def stretch_key(cls, hex_seed: str) -> int:
+        assert is_hex_str(hex_seed), f"expected hex str, got {type(hex_seed)} with {len(hex_seed)=}"
+        encoded_hex_seed = hex_seed.encode('ascii')
+        x = encoded_hex_seed
         for i in range(100000):
-            x = hashlib.sha256(x + seed).digest()
+            x = hashlib.sha256(x + encoded_hex_seed).digest()
         return string_to_number(x)
 
     @classmethod
-    def get_sequence(self, mpk, for_change, n):
+    def get_sequence(cls, mpk: str, for_change: int, n: int) -> int:
         return string_to_number(sha256d(("%d:%d:"%(n, for_change)).encode('ascii') + bfh(mpk)))
 
     @classmethod
-    def get_pubkey_from_mpk(cls, mpk, for_change, n) -> bytes:
+    def get_pubkey_from_mpk(cls, mpk: str, for_change: int, n: int) -> bytes:
         z = cls.get_sequence(mpk, for_change, n)
         master_public_key = ecc.ECPubkey(bfh('04'+mpk))
         public_key = master_public_key + z*ecc.GENERATOR
@@ -750,22 +810,24 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
             raise CannotDerivePubkey("forbidden path")
         return self.get_pubkey_from_mpk(self.mpk, for_change, n)
 
-    def _get_private_key_from_stretched_exponent(self, for_change, n, secexp):
+    def _get_private_key_from_stretched_exponent(self, for_change: int, n: int, secexp: int) -> bytes:
         secexp = (secexp + self.get_sequence(self.mpk, for_change, n)) % ecc.CURVE_ORDER
         pk = int.to_bytes(secexp, length=32, byteorder='big', signed=False)
         return pk
 
     def get_private_key(self, sequence: Sequence[int], password):
-        seed = self.get_hex_seed(password)
-        secexp = self.stretch_key(seed)
-        self._check_seed(seed, secexp=secexp)
+        hex_seed = self._get_hex_seed(password)
+        secexp = self.stretch_key(hex_seed)
+        self._check_seed(hex_seed, secexp=secexp)
         for_change, n = sequence
+        assert isinstance(for_change, int), type(for_change)
+        assert isinstance(n, int), type(n)
         pk = self._get_private_key_from_stretched_exponent(for_change, n, secexp)
         return pk, False
 
-    def _check_seed(self, seed, *, secexp=None):
+    def _check_seed(self, hex_seed: str, *, secexp: int = None) -> None:
         if secexp is None:
-            secexp = self.stretch_key(seed)
+            secexp = self.stretch_key(hex_seed)
         master_private_key = ecc.ECPrivkey.from_secret_scalar(secexp)
         master_public_key = master_private_key.get_public_key_bytes(compressed=False)[1:]
         if master_public_key != bfh(self.mpk):
@@ -773,8 +835,8 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
 
     @also_test_none_password
     def check_password(self, password):
-        seed = self.get_hex_seed(password)
-        self._check_seed(seed)
+        hex_seed = self._get_hex_seed(password)
+        self._check_seed(hex_seed)
 
     def get_master_public_key(self):
         return self.mpk
@@ -798,10 +860,17 @@ class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):
         fingerprint_hex = self.get_root_fingerprint()
         der_prefix_str = self.get_derivation_prefix()
         fingerprint_bytes = bfh(fingerprint_hex)
-        der_prefix_ints = convert_bip32_path_to_list_of_uint32(der_prefix_str)
+        der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
         der_full = der_prefix_ints + list(der_suffix)
         return fingerprint_bytes, der_full
 
+    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
+        return PubkeyProvider(
+            origin=None,
+            pubkey=self.derive_pubkey(*sequence).hex(),
+            deriv_path=None,
+        )
+
     def update_password(self, old_password, new_password):
         self.check_password(old_password)
         if new_password == '':
@@ -832,7 +901,14 @@ class Hardware_KeyStore(Xpub, KeyStore):
         self.handler = None  # type: Optional[HardwareHandlerBase]
         run_hook('init_keystore', self)
 
-    def set_label(self, label):
+    def watching_only_keystore(self):
+        return BIP32_KeyStore({
+            'xpub': self.xpub,
+            'root_fingerprint': self.get_root_fingerprint(),
+            'derivation': self.get_derivation_prefix(),
+        })
+
+    def set_label(self, label: Optional[str]) -> None:
         self.label = label
 
     def may_have_password(self):
@@ -851,23 +927,13 @@ class Hardware_KeyStore(Xpub, KeyStore):
             'xpub': self.xpub,
             'derivation': self.get_derivation_prefix(),
             'root_fingerprint': self.get_root_fingerprint(),
-            'label':self.label,
+            'label': self.label,
             'soft_device_id': self.soft_device_id,
         }
 
-    def unpaired(self):
-        '''A device paired with the wallet was disconnected.  This can be
-        called in any thread context.'''
-        self.logger.info("unpaired")
-
-    def paired(self):
-        '''A device paired with the wallet was (re-)connected.  This can be
-        called in any thread context.'''
-        self.logger.info("paired")
-
     def is_watching_only(self):
-        '''The wallet is not watching-only; the user will be prompted for
-        pin and passphrase as appropriate when needed.'''
+        """The wallet is not watching-only; the user will be prompted for
+        pin and passphrase as appropriate when needed."""
         assert not self.has_seed()
         return False
 
@@ -915,16 +981,24 @@ class Hardware_KeyStore(Xpub, KeyStore):
             self.soft_device_id = client.get_soft_device_id()
             self.is_requesting_to_be_rewritten_to_wallet_file = True
 
+    def pairing_code(self) -> Optional[str]:
+        """Used by the DeviceMgr to keep track of paired hw devices."""
+        if not self.soft_device_id:
+            return None
+        return f"{self.plugin.name}/{self.soft_device_id}"
+
 
 KeyStoreWithMPK = Union[KeyStore, MasterPublicKeyMixin]  # intersection really...
 AddressIndexGeneric = Union[Sequence[int], str]  # can be hex pubkey str
 
 
-def bip39_normalize_passphrase(passphrase):
+def bip39_normalize_passphrase(passphrase: str):
     return normalize('NFKD', passphrase or '')
 
-def bip39_to_seed(mnemonic, passphrase):
-    import hashlib, hmac
+
+def bip39_to_seed(mnemonic: str, *, passphrase: Optional[str]) -> bytes:
+    import hashlib
+    passphrase = passphrase or ""
     PBKDF2_ROUNDS = 2048
     mnemonic = normalize('NFKD', ' '.join(mnemonic.split()))
     passphrase = bip39_normalize_passphrase(passphrase)
@@ -966,11 +1040,16 @@ def bip39_is_checksum_valid(
     return checksum == calculated_checksum, True
 
 
-def from_bip43_rootseed(root_seed, derivation, xtype=None):
+def from_bip43_rootseed(
+    root_seed: bytes,
+    *,
+    derivation: str,
+    xtype: Optional[str] = None,
+):
     k = BIP32_KeyStore({})
     if xtype is None:
         xtype = xtype_from_derivation(derivation)
-    k.add_xprv_from_seed(root_seed, xtype, derivation)
+    k.add_xprv_from_seed(root_seed, xtype=xtype, derivation=derivation)
     return k
 
 
@@ -983,7 +1062,7 @@ PURPOSE48_SCRIPT_TYPES_INV = inv_dict(PURPOSE48_SCRIPT_TYPES)
 
 def xtype_from_derivation(derivation: str) -> str:
     """Returns the script type to be used for this derivation."""
-    bip32_indices = convert_bip32_path_to_list_of_uint32(derivation)
+    bip32_indices = convert_bip32_strpath_to_intpath(derivation)
     if len(bip32_indices) >= 1:
         if bip32_indices[0] == 84 + BIP32_PRIME:
             return 'p2wpkh'
@@ -1004,9 +1083,9 @@ def xtype_from_derivation(derivation: str) -> str:
     return 'standard'
 
 
-hw_keystores = {}
+hw_keystores = {}  # type: Dict[str, Type[Hardware_KeyStore]]
 
-def register_keystore(hw_type, constructor):
+def register_keystore(hw_type: str, constructor: Type[Hardware_KeyStore]) -> None:
     hw_keystores[hw_type] = constructor
 
 def hardware_keystore(d) -> Hardware_KeyStore:
@@ -1018,7 +1097,9 @@ def hardware_keystore(d) -> Hardware_KeyStore:
                               f'hw_keystores: {list(hw_keystores)}')
 
 def load_keystore(db: 'WalletDB', name: str) -> KeyStore:
-    d = db.get(name, {})
+    # deepcopy object to avoid keeping a pointer to db.data
+    # note: this is needed as type(wallet.db.get("keystore")) != StoredDict
+    d = copy.deepcopy(db.get(name, {}))
     t = d.get('type')
     if not t:
         raise WalletFileException(
@@ -1037,23 +1118,23 @@ def load_keystore(db: 'WalletDB', name: str) -> KeyStore:
 def is_old_mpk(mpk: str) -> bool:
     try:
         int(mpk, 16)  # test if hex string
-    except:
+    except Exception:
         return False
     if len(mpk) != 128:
         return False
     try:
         ecc.ECPubkey(bfh('04' + mpk))
-    except:
+    except Exception:
         return False
     return True
 
 
-def is_address_list(text):
+def is_address_list(text: str) -> bool:
     parts = text.split()
     return bool(parts) and all(bitcoin.is_address(x) for x in parts)
 
 
-def get_private_keys(text, *, allow_spaces_inside_key=True, raise_on_error=False):
+def get_private_keys(text: str, *, allow_spaces_inside_key=True, raise_on_error=False) -> Optional[Sequence[str]]:
     if allow_spaces_inside_key:  # see #1612
         parts = text.split('\n')
         parts = map(lambda x: ''.join(x.split()), parts)
@@ -1062,23 +1143,24 @@ def get_private_keys(text, *, allow_spaces_inside_key=True, raise_on_error=False
         parts = text.split()
     if bool(parts) and all(bitcoin.is_private_key(x, raise_on_error=raise_on_error) for x in parts):
         return parts
+    return None
 
 
-def is_private_key_list(text, *, allow_spaces_inside_key=True, raise_on_error=False):
+def is_private_key_list(text: str, *, allow_spaces_inside_key: bool = True, raise_on_error: bool = False) -> bool:
     return bool(get_private_keys(text,
                                  allow_spaces_inside_key=allow_spaces_inside_key,
                                  raise_on_error=raise_on_error))
 
 
-def is_master_key(x):
+def is_master_key(x: str) -> bool:
     return is_old_mpk(x) or is_bip32_key(x)
 
 
-def is_bip32_key(x):
+def is_bip32_key(x: str) -> bool:
     return is_xprv(x) or is_xpub(x)
 
 
-def bip44_derivation(account_id, bip43_purpose=44):
+def bip44_derivation(account_id: int, bip43_purpose: int = 44) -> str:
     coin = constants.net.BIP44_COIN_TYPE
     der = "m/%d'/%d'/%d'" % (bip43_purpose, coin, int(account_id))
     return normalize_bip32_derivation(der)
@@ -1096,49 +1178,52 @@ def purpose48_derivation(account_id: int, xtype: str) -> str:
     return normalize_bip32_derivation(der)
 
 
-def from_seed(seed, passphrase, is_p2sh=False):
-    t = seed_type(seed)
+def from_seed(seed: str, *, passphrase: Optional[str], for_multisig: bool = False) -> Union[BIP32_KeyStore, Old_KeyStore]:
+    passphrase = passphrase or ""
+    t = calc_seed_type(seed)
     if t == 'old':
+        if passphrase:
+            raise Exception("'old'-type electrum seed cannot have passphrase")
         keystore = Old_KeyStore({})
         keystore.add_seed(seed)
     elif t in ['standard', 'segwit']:
         keystore = BIP32_KeyStore({})
         keystore.add_seed(seed)
         keystore.passphrase = passphrase
-        bip32_seed = Mnemonic.mnemonic_to_seed(seed, passphrase)
+        bip32_seed = Mnemonic.mnemonic_to_seed(seed, passphrase=passphrase)
         if t == 'standard':
             der = "m/"
             xtype = 'standard'
         else:
-            der = "m/1'/" if is_p2sh else "m/0'/"
-            xtype = 'p2wsh' if is_p2sh else 'p2wpkh'
-        keystore.add_xprv_from_seed(bip32_seed, xtype, der)
+            der = "m/1'/" if for_multisig else "m/0'/"
+            xtype = 'p2wsh' if for_multisig else 'p2wpkh'
+        keystore.add_xprv_from_seed(bip32_seed, xtype=xtype, derivation=der)
     else:
         raise BitcoinException('Unexpected seed type {}'.format(repr(t)))
     return keystore
 
-def from_private_key_list(text):
+def from_private_key_list(text: str) -> Imported_KeyStore:
     keystore = Imported_KeyStore({})
     for x in get_private_keys(text):
         keystore.import_privkey(x, None)
     return keystore
 
-def from_old_mpk(mpk):
+def from_old_mpk(mpk: str) -> Old_KeyStore:
     keystore = Old_KeyStore({})
     keystore.add_master_public_key(mpk)
     return keystore
 
-def from_xpub(xpub):
+def from_xpub(xpub: str) -> BIP32_KeyStore:
     k = BIP32_KeyStore({})
     k.add_xpub(xpub)
     return k
 
-def from_xprv(xprv):
+def from_xprv(xprv: str) -> BIP32_KeyStore:
     k = BIP32_KeyStore({})
     k.add_xprv(xprv)
     return k
 
-def from_master_key(text):
+def from_master_key(text: str) -> Union[BIP32_KeyStore, Old_KeyStore]:
     if is_xprv(text):
         k = from_xprv(text)
     elif is_old_mpk(text):
