@@ -8,10 +8,13 @@ https://github.com/lightningnetwork/lightning-rfc/blob/master/09-features.md
 import asyncio
 import os
 import time
+from typing import Optional
+
+from aiorpcx import NetAddress
 
 from electrum.logging import get_logger, configure_logging
 from electrum.simple_config import SimpleConfig
-from electrum import constants
+from electrum import constants, util
 from electrum.daemon import Daemon
 from electrum.wallet import create_new_wallet
 from electrum.util import create_and_start_event_loop, log_exceptions, bfh
@@ -39,7 +42,7 @@ loop, stopping_fut, loop_thread = create_and_start_event_loop()
 time.sleep(2)
 
 if IS_TESTNET:
-    constants.set_testnet()
+    constants.BitcoinTestnet.set_as_network()
 daemon = Daemon(config, listen_jsonrpc=False)
 network = daemon.network
 assert network.asyncio_loop.is_running()
@@ -51,8 +54,7 @@ if not os.path.exists(wallet_path):
     create_new_wallet(path=wallet_path, config=config)
 
 # open wallet
-wallet = daemon.load_wallet(wallet_path, password=None, manual_upgrades=False)
-wallet.start_network(network)
+wallet = daemon.load_wallet(wallet_path, password=None, upgrade=True)
 
 
 async def worker(work_queue: asyncio.Queue, results_queue: asyncio.Queue, flag):
@@ -67,24 +69,20 @@ async def worker(work_queue: asyncio.Queue, results_queue: asyncio.Queue, flag):
         work = await work_queue.get()
 
         # only check non-onion addresses
-        addr = None
-        for a in work['addrs']:
-            if "onion" not in a[0]:
+        addr = None  # type: Optional[NetAddress]
+        for a in work['addrs']:  # type: NetAddress
+            if not str(a.host).endswith(".onion"):
                 addr = a
         if not addr:
             await results_queue.put(None)
             continue
 
-        # handle ipv4/ipv6
-        if ':' in addr[0]:
-            connect_str = f"{work['pk'].hex()}@[{addr.host}]:{addr.port}"
-        else:
-            connect_str = f"{work['pk'].hex()}@{addr.host}:{addr.port}"
+        connect_str = f"{work['pk'].hex()}@{addr}"
 
         print(f"worker connecting to {connect_str}")
         try:
             peer = await wallet.lnworker.add_peer(connect_str)
-            res = await asyncio.wait_for(peer.initialized, TIMEOUT)
+            res = await util.wait_for2(peer.initialized, TIMEOUT)
             if res:
                 if peer.features & flag == work['features'] & flag:
                     await results_queue.put(True)
@@ -109,7 +107,7 @@ async def node_flag_stats(opt_flag: LnFeatures, presync: False):
     try:
         await wallet.lnworker.channel_db.data_loaded.wait()
 
-        # optionally presync graph (not relyable)
+        # optionally presync graph (not reliable)
         if presync:
             network.start_gossip()
 
@@ -126,7 +124,7 @@ async def node_flag_stats(opt_flag: LnFeatures, presync: False):
         with wallet.lnworker.channel_db.lock:
             nodes = wallet.lnworker.channel_db._nodes.copy()
 
-        # check how many nodes advertize opt/req flag in the gossip
+        # check how many nodes advertise opt/req flag in the gossip
         n_opt = 0
         n_req = 0
         print(f"analyzing {len(nodes.keys())} nodes")
@@ -177,3 +175,5 @@ async def node_flag_stats(opt_flag: LnFeatures, presync: False):
 
 asyncio.run_coroutine_threadsafe(
     node_flag_stats(FLAG, presync=PRESYNC), loop)
+while loop_thread.is_alive():
+    loop_thread.join(1)
